#include <ctype.h>
#include "scene.h"
#include "adjacency.h"
#include "adjacency3.h"
#include "common.h"

Scene  *CreateScene(int xsize,int ysize,int zsize)
{
  Scene *scn=NULL;
  int i,xysize;

  scn = (Scene *) calloc(1,sizeof(Scene));
  if (scn == NULL){
    Error(MSG1,"CreateScene");
  }
  
  scn->data    = AllocIntArray(xsize*ysize*zsize);
  scn->xsize   = xsize;
  scn->ysize   = ysize;
  scn->zsize   = zsize;
  scn->dx      = 1.0;
  scn->dy      = 1.0;
  scn->dz      = 1.0;
  scn->tby     = AllocIntArray(ysize);
  scn->tbz     = AllocIntArray(zsize);
  scn->maxval  = 0;
  scn->n       = xsize*ysize*zsize;

  if (scn->data==NULL || scn->tbz==NULL || scn->tby==NULL) {
    Error(MSG1,"CreateScene");
  }

  scn->tby[0]=0;
  for (i=1; i < ysize; i++)
    scn->tby[i]=scn->tby[i-1] + xsize;

  scn->tbz[0]=0; xysize = xsize*ysize;
  for (i=1; i < zsize; i++)
    scn->tbz[i]=scn->tbz[i-1] + xysize;

  return(scn);
}

Scene *CopyScene(Scene *scn){
  Scene *aux = CreateScene(scn->xsize,scn->ysize,scn->zsize);
  aux->dx     = scn->dx;
  aux->dy     = scn->dy;
  aux->dz     = scn->dz;
  aux->maxval = scn->maxval;

  aux->n=scn->xsize*scn->ysize*scn->zsize;
  memcpy(aux->data,scn->data,sizeof(int) * aux->n);
  return aux;
}

void     DestroyScene(Scene **scn)
{
  Scene *aux;

  aux = *scn;
  if(aux != NULL){
    if (aux->data != NULL)  free(aux->data); 
    if (aux->tby  != NULL)  free(aux->tby);
    if (aux->tbz  != NULL)  free(aux->tbz);
    free(aux);    
    *scn = NULL;
  }
}

bool ValidVoxel(Scene *scn, int vx, int vy, int vz)
{
  if ((vx >= 0)&&(vx < scn->xsize)&&
      (vy >= 0)&&(vy < scn->ysize)&&
      (vz >= 0)&&(vz < scn->zsize))
    return(true);
  else
    return(false);
}
Voxel   Transform_Voxel(float M[4][4],Voxel v){
  Voxel nv;

  nv.x = (int) (M[0][0]*v.x + M[0][1]*v.y + M[0][2]*v.z + M[0][3]);
  nv.y = (int) (M[1][0]*v.x + M[1][1]*v.y + M[1][2]*v.z + M[1][3]);
  nv.z = (int) (M[2][0]*v.x + M[2][1]*v.y + M[2][2]*v.z + M[2][3]);

  return(nv);

}

int MaximumValue3(Scene *scn)
{
  unsigned int i, n, r;
  int Imax;
      
  Imax = 0;
  n = scn->xsize*scn->ysize*scn->zsize - 1;
  r = n%4;
  n -= r;
  for (i=0; i < n; i+=4) {
    if (scn->data[i] > Imax)
      Imax = scn->data[i];
    if (scn->data[i+1] > Imax)
      Imax = scn->data[i+1];
    if (scn->data[i+2] > Imax)
      Imax = scn->data[i+2];
    if (scn->data[i+3] > Imax)
      Imax = scn->data[i+3];
  }
  while (r != 0) {
    if (scn->data[i+r-1] > Imax)
      Imax = scn->data[i+r-1];
    --r;
  }

  scn->maxval = Imax;

  return(Imax); 
}

void SetSceneImax(Scene *scn, int val) {
  scn->maxval = val;
}

int SceneImax(Scene *scn)
{
  return scn->maxval;
}

int MinimumValue3(Scene *scn)
{
  int i,n;
  int Imin;
  
  Imin = 65535; 
  n = scn->xsize*scn->ysize*scn->zsize;
  for (i=0; i < n; i++) 
    if (scn->data[i] < Imin)
      Imin = scn->data[i];
  
  return(Imin); 
}

#ifdef _MSC_VER
#define strcasecmp stricmp
#endif

Scene *ReadScene(char *filename)
{
  Scene  *scn=NULL;	
  FILE   *fp=NULL;
  uchar  *data8=NULL;
#if _WIN32 || __BYTE_ORDER==__LITTLE_ENDIAN 
  ushort *data16=NULL;
#endif
  char    type[10];
  int     i,n,v,xsize,ysize,zsize;
  long    pos;
  int     null;

  // Checking file type
  int len = strlen(filename);
  if ( (len<=4) || (strcasecmp(filename + len - 4, ".scn")!=0)) {
    Error(MSG2,"ReadScene: Invalid file name or extension.");
    return NULL;
  }

  // Read the scn file
  fp = fopen(filename,"rb");
  if (fp == NULL){
    Error(MSG2,"ReadScene");
  }
  null = fscanf(fp,"%s\n",type);
  if((strcmp(type,"SCN")==0)){
    null = fscanf(fp,"%d %d %d\n",&xsize,&ysize,&zsize);
    scn = CreateScene(xsize,ysize,zsize);
    n = xsize*ysize*zsize;
    null = fscanf(fp,"%f %f %f\n",&scn->dx,&scn->dy,&scn->dz);
    null = fscanf(fp,"%d",&v);
    pos = ftell(fp);
    fseek(fp,(pos+1)*sizeof(char),SEEK_SET); // +1 for the EOL \n character not included in last fscanf() call
    if (v==8){			
      data8  = AllocUCharArray(n);			
      null = fread(data8,sizeof(uchar),n,fp);
      for (i=0; i < n; i++) 
	scn->data[i] = (int) data8[i];
      free(data8);
    } else if (v==16) {
      // assuming that data was written with LITTLE-ENDIAN Byte ordering, i.e. LSB...MSB
#if _WIN32 || __BYTE_ORDER==__LITTLE_ENDIAN
      // for PCs, Intel microprocessors (LITTLE-ENDIAN too) -> no change
      data16 = AllocUShortArray(n);
      null = fread(data16,sizeof(ushort),n,fp);
      for (i=0; i < n; i++)
	scn->data[i] = (int) data16[i];
      free(data16);
#else
      // for Motorola, IBM, SUN (BIG-ENDIAN) -> SWAp Bytes!
      Warning("Data is converted from LITTLE to BIG-ENDIAN","ReadScene");
      data8 = AllocUCharArray(2*n);
      null = fread(data8,sizeof(uchar),2*n,fp);
      j=0;
      for (i=0; i < 2*n; i+=2) {
	scn->data[j] = (int) data8[i] + 256 * (int) data8[i+1];
	j++;
      }
      free(data8);
#endif
    } else { /* n = 32 */
      //Warning("32-bit data. Values may be wrong (little/big endian byte ordering)","ReadScene");
      n = xsize*ysize*zsize;
      null = fread(scn->data,sizeof(int),n,fp);      
    }
    fclose(fp);
  } else {
    fprintf(stderr,"Input scene must be SCN\n");
    exit(-1);
  }
  scn->maxval = MaximumValue3(scn);
  return(scn);
}

void WriteScene(Scene *scn, char *filename) 
{
  FILE *fp=NULL;
  int Imax;
  int i,n;
  uchar  *data8 =NULL;
  ushort *data16=NULL;

  // Checking file type
  int len = strlen(filename);
  if ( (len<=4) || (strcasecmp(filename + len - 4, ".scn")!=0)) {
    Error(MSG2,"WriteScene: Invalid file name or extension.");
  }

  // Writing the scn file
  fp = fopen(filename,"wb"); 
  if (fp == NULL) 
    Error(MSG2,"WriteScene");

  fprintf(fp,"SCN\n");
  fprintf(fp,"%d %d %d\n",scn->xsize,scn->ysize,scn->zsize);
  fprintf(fp,"%f %f %f\n",scn->dx,scn->dy,scn->dz);
  
  Imax = MaximumValue3(scn);
  
  n = scn->xsize*scn->ysize*scn->zsize;
  if (Imax < 256) {
    fprintf(fp,"%d\n",8);
    data8 = AllocUCharArray(n);
    for (i=0; i < n; i++) 
      data8[i] = (uchar) scn->data[i];
    fwrite(data8,sizeof(uchar),n,fp);
    free(data8);
  } else if (Imax < 65536) {
    fprintf(fp,"%d\n",16);
    data16 = AllocUShortArray(n);
    for (i=0; i < n; i++)
      data16[i] = (ushort) scn->data[i];
    fwrite(data16,sizeof(ushort),n,fp);
    free(data16);
  } else {
    fprintf(fp,"%d\n",32);
    fwrite(scn->data,sizeof(int),n,fp);
  }
  fclose(fp);
}

int VoxelValue(Scene *scn,Voxel v){
  int k,aux;
  if(ValidVoxel(scn,v.x,v.y,v.z)){
  k=v.x+scn->tby[v.y]+scn->tbz[v.z];
  aux=scn->data[k];
  } else aux=0;
  
  return aux;
}

Scene *LinearInterp(Scene *scn,float dx,float dy,float dz){
  int value;
  Scene *scene,*tmp;
  Voxel P,Q,R; /* previous, current, and next voxel */
  float min=(float)INT_MAX;
  float walked_dist,dist_PQ;

  /* The default voxel sizes of the input scene should be dx=dy=dz=1.0 */

  if ((scn->dx == 0.0) && (scn->dy == 0.0) && (scn->dz == 0.0)) {    
  scn->dx=1.0;
  scn->dy=1.0;
  scn->dz=1.0;
  }

  /* The default voxel sizes of the output scene should be dx=dy=dz=min(dx,dy,dz) */

  if ((dx == 0.0) || (dy == 0.0) || (dz == 0.0)) {
    if (scn->dx < min)
      min = scn->dx;
    if (scn->dy < min)
      min = scn->dy;
    if (scn->dz < min)
      min = scn->dz;
    dx = min; dy = min; dz = min;
    if (min <= 0) {
      fprintf(stderr,"Voxel distance can not be negative.\n");
      exit(-1);
    }
  }

  /* If there is no need for resampling then returns input scene */

  if ((dx == scn->dx) && (dy == scn->dy) && (dz == scn->dz)) {
    scene = CopyScene(scn);
    return (scene);
  } else {
  /* Else the working image is the input image */
    scene = scn;
  }

  /* Resample in x */

  if (dx != scn->dx) {
    tmp = CreateScene(ROUND((float)(scene->xsize)*scene->dx/dx),scene->ysize,scene->zsize);
    for(Q.x=0; Q.x < tmp->xsize; Q.x++)
      for(Q.z=0; Q.z < tmp->zsize; Q.z++)
        for(Q.y=0; Q.y < tmp->ysize; Q.y++) {


            walked_dist = (float)Q.x * dx; /* the walked distance so far */

            P.x = (int)(walked_dist/scn->dx); /* P is the previous pixel in the
                                             original scene */
            P.y = Q.y;
            P.z = Q.z;

            R.x = P.x + 1; /* R is the next pixel in the original
                              image. Observe that Q is in between P
                              and R. */
            R.y = P.y;
            R.z = P.z;

            dist_PQ =  walked_dist - (float)P.x * scn->dx;  /* the distance between P and Q */

            /* interpolation: P --- dPQ --- Q ---- dPR-dPQ ---- R

               I(Q) = (I(P)*(dPR-dPQ) + I(R)*dPQ) / dPR

            */

            value = ROUND((( scn->dx - dist_PQ)*(float)VoxelValue(scene,P) + dist_PQ * (float)VoxelValue(scene,R) )/scn->dx);
            tmp->data[Q.x+tmp->tby[Q.y]+tmp->tbz[Q.z]]=value;
	  }
    scene=tmp;
  }

  /* Resample in y */

  if (dy != scn->dy) {
    tmp = CreateScene(scene->xsize, ROUND((float)scene->ysize * scn->dy / dy),scene->zsize);
    for(Q.y=0; Q.y < tmp->ysize; Q.y++)
      for(Q.z=0; Q.z < tmp->zsize; Q.z++)
          for(Q.x=0; Q.x < tmp->xsize; Q.x++) {

            walked_dist = (float)Q.y * dy;

            P.x = Q.x;
            P.y = (int)(walked_dist/scn->dy);
            P.z = Q.z;

            R.x = P.x;
            R.y = P.y + 1;
            R.z = P.z;

            dist_PQ =  walked_dist - (float)P.y * scn->dy;
	    /* comecar a adaptar daqui !! */
            value = ROUND((( (scn->dy - dist_PQ)*(float)VoxelValue(scene,P)) + dist_PQ * (float)VoxelValue(scene,R)) / scn->dy) ;
	    tmp->data[Q.x+tmp->tby[Q.y]+tmp->tbz[Q.z]]=value;
           
          }
    if (scene != scn) {
      DestroyScene(&scene);
    }
    scene=tmp;
  }

  /* Resample in z */

  if (dz != scn->dz) {
    tmp = CreateScene(scene->xsize,scene->ysize,ROUND((float)scene->zsize * scn->dz / dz));
    for(Q.z=0; Q.z < tmp->zsize; Q.z++)
        for(Q.y=0; Q.y < tmp->ysize; Q.y++)
          for(Q.x=0; Q.x < tmp->xsize; Q.x++) {

            walked_dist = (float)Q.z * dz;

            P.x = Q.x;
            P.y = Q.y;
            P.z = (int)(walked_dist/scn->dz);

            R.x = P.x;
            R.y = P.y;
            R.z = P.z + 1;

            dist_PQ =  walked_dist - (float)P.z * scn->dz;

	    value = ROUND((( (scn->dz - dist_PQ)*(float)VoxelValue(scene,P)) + dist_PQ * (float)VoxelValue(scene,R)) / scn->dz) ;
	    tmp->data[Q.x+tmp->tby[Q.y]+tmp->tbz[Q.z]]=value;
	  }
    if (scene != scn) {
      DestroyScene(&scene);
    }
    scene=tmp;
  }
 
  scene->dx=dx;
  scene->dy=dy;
  scene->dz=dz;
  scene->maxval=scn->maxval;
  return(scene);
}

Scene *KnnInterp(Scene *scn,float dx,float dy,float dz){
  Scene *new_scn;
  float min, x, y, z;
  Voxel u;

  /* The voxel sizes default in the input scene should be
     dx=dy=dz=1.0 */

  if ((scn->dx == 0.0) && (scn->dy == 0.0) && (scn->dz == 0.0)) {    
    scn->dx=1.0;
    scn->dy=1.0;
    scn->dz=1.0;
  }

  /* The voxel sizes default in the output scene should be
     dx=dy=dz=min(dx,dy,dz) */

  if ((dx == 0.0) || (dy == 0.0) || (dz == 0.0)) {
    min = FLT_MAX;
    if (scn->dx < min)
      min = scn->dx;
    if (scn->dy < min)
      min = scn->dy;
    if (scn->dz < min)
      min = scn->dz;
    dx = min; dy = min; dz = min;
    if (min <= 0.0) {
      fprintf(stderr,"Voxel distance cannot be negative.\n");
      exit(-1);
    }
  }

  /* If there is no need for resampling then return the input scene */

  if ((dx == scn->dx) && (dy == scn->dy) && (dz == scn->dz)) {
    new_scn = CopyScene(scn);
  } else {
    /* do knn interpolation */
    new_scn = CreateScene(ROUND((float)(scn->xsize*scn->dx/dx)),ROUND((float)(scn->ysize*scn->dy/dy)),ROUND((float)(scn->zsize-1)*scn->dz/dz));
    new_scn->dx     = dx;
    new_scn->dy     = dy;
    new_scn->dz     = dz;
    dx              = dx / scn->dx;
    dy              = dy / scn->dy;
    dz              = dz / scn->dz;
    new_scn->maxval = scn->maxval;
    for (z=0.0,u.z=0; z < scn->zsize; z=z+dz, u.z++)
      for (y=0.0, u.y=0; y < scn->ysize; y=y+dy, u.y++)
	for (x=0.0, u.x=0; x < scn->xsize; x=x+dx, u.x++){
	  
	  if ((x==(int)x)&&(y==(int)y)&&(z==(int)z)){
	    new_scn->data[u.x + new_scn->tby[u.y] + new_scn->tbz[u.z]]=
	      scn->data[(int)x + scn->tby[(int)y] + scn->tbz[(int)z]];
	  }else{
	    new_scn->data[u.x + new_scn->tby[u.y] + new_scn->tbz[u.z]]=KnnInterpValue(scn, x, y, z); 	    
	  }
	}
  }
  return(new_scn);
}

#define KNNSIZE 7

int KnnInterpValue(Scene *scn, float x, float y, float z)
{
  int j;
  Voxel u, v;
  Voxel vox[KNNSIZE];
  int value=scn->data[(int)x+scn->tby[(int)y]+scn->tbz[(int)z]];
  int absdif[KNNSIZE];
  int d;
  int i,q;
  float new_value, w, cw;

  for (i=0; i < KNNSIZE; i++) 
    absdif[i]=INT_MAX;

  j=0;
  for (u.z = (int)z-1; u.z <= (int)z+1; u.z++) 
    for (u.y = (int)y-1; u.y <= (int)y+1; u.y++) 
      for (u.x = (int)x-1; u.x <= (int)x+1; u.x++) {

	if (ValidVoxel(scn,u.x,u.y,u.z)){
	  j++;
	  q           = u.x+scn->tby[u.y]+scn->tbz[u.z];
	  i           = KNNSIZE-1;
	  absdif[i]   = abs(scn->data[q]-value);
	  vox[i]    = u;

	  // find the nearest neighbors
	  while ((i > 0)&&(absdif[i]<absdif[i-1])) {
	    d           = absdif[i];
	    v           = vox[i];
	    absdif[i]   = absdif[i-1];
	    vox[i]    = vox[i-1];
	    absdif[i-1] = d;
	    vox[i-1]  = v;
	    i--;
	  }
	}

      }

  if (j < KNNSIZE) return((int)value);

  // compute interpolation
  new_value = 0.0; cw = 0.0;
  for (i=0; i < KNNSIZE-1; i++) {
    // w = (vox[i].x - x)*(vox[i].x - x) + 
    //  (vox[i].y - y)*(vox[i].y - y) + 
    //  (vox[i].z - z)*(vox[i].z - z);
    //  w = exp(-w/4.0);
    //  w = 1.0/(absdif[i]+1.0);
    w = KNNSIZE-i;
    cw += w;
    new_value += scn->data[vox[i].x+scn->tby[vox[i].y] + scn->tbz[vox[i].z]]*w;
  }

  return((int)(new_value/cw));
}
 	    
int GetXSize(Scene *scn) {
  return scn->xsize;
}

int GetYSize(Scene *scn) {
  return scn->ysize;
}

int GetZSize(Scene *scn) {
  return scn->zsize;
}

float GetDx(Scene *scn) {
  return scn->dx;
}

float GetDy(Scene *scn) {
  return scn->dy;
}

float GetDz(Scene *scn) {
  return scn->dz;
}

void SetDx(Scene *scn, float dx) {
  scn->dx = dx;
}

void SetDy(Scene *scn, float dy) {
  scn->dy = dy;
}

void SetDz(Scene *scn, float dz) {
  scn->dz = dz;
}

void SetVoxelSize(Scene *scn, float dx, float dy, float dz){
  SetDx(scn, dx);
  SetDy(scn, dy);
  SetDz(scn, dz);
}

int Coord2Voxel(Scene *scn, int x, int y, int z) {
  return (x + scn->tby[y] + scn->tbz[z]);
}

int GetVoxelValue(Scene *scn, int voxel) {
  return (scn->data[voxel]);
}

int ScenesAreEqual(Scene *scn1, Scene *scn2)
{
  int i,n;

  if((scn1->xsize != scn2->xsize)||(scn1->ysize != scn2->ysize)||(scn1->zsize != scn2->zsize)) {
    fprintf(stderr,"Scenes have different sizes\n");
    return(0);
  }
  n = scn1->xsize * scn1->ysize * scn1->zsize;
  for(i=0;i<n;i++) {
    if(scn1->data[i] != scn2->data[i]) {
      fprintf(stderr,"Scenes have different values (in voxel %d for example)\n",i);
      return(0);
    }
  }
  return(1);
}

void CompareScenes(Scene *scn1, Scene *scn2)
{
  int i,n,nvoxel=0;
  //  int diff;
  //  long sum=0;
  //  float mdiff;

  if((scn1->xsize != scn2->xsize)||(scn1->ysize != scn2->ysize)||(scn1->zsize != scn2->zsize)) {
    fprintf(stderr,"Scenes have different sizes\n");
    return;
  }
  n = scn1->xsize * scn1->ysize * scn1->zsize;
  for(i=0;i<n;i++) {
    if(scn1->data[i] != scn2->data[i]) {
      //  fprintf(stderr,"Scenes have different values (%d and %d) in voxel %d\n",scn1->data[i],scn2->data[i],i);
      //      diff = scn1->data[i] - scn2->data[i];
      //      if(diff<0) diff = -diff;
      //      sum += diff;
      nvoxel++;
    }
  }
  //  mdiff = (float) sum/nvoxel;
  printf("%d voxels are different in a scene of %d voxels\n",nvoxel,n);
  //  printf("the mean difference for different voxels is %f\n", mdiff);
  return;
}

